@@ -7,7 +7,6 @@
 
 #include "state.h"
 #include "screen.h"
-#include "lineops.h"
 #include "charsets.h"
 #include <limits.h>
 #include <math.h>
@@ -17,7 +16,7 @@
 
 extern PyTypeObject Screen_Type;
 
-static MouseShape mouse_cursor_shape = BEAM;
+static MouseShape mouse_cursor_shape = TEXT_POINTER;
 typedef enum MouseActions { PRESS, RELEASE, DRAG, MOVE } MouseAction;
 #define debug(...) if (OPT(debug_keyboard)) printf(__VA_ARGS__);
 
@@ -67,11 +66,11 @@ static char mouse_event_buf[64];
 
 static int
 encode_mouse_event_impl(const MousePosition *mpos, int mouse_tracking_protocol, int button, MouseAction action, int mods) {
-    unsigned int cb = 0;
+    unsigned int cb = encode_button(button);
     if (action == MOVE) {
-        cb = 3;
+        if (cb == UINT_MAX) cb = 3;
+        cb += 32;
     } else {
-        cb = encode_button(button);
         if (cb == UINT_MAX) return 0;
     }
     if (action == DRAG || action == MOVE) cb |= MOTION_INDICATOR;
@@ -120,9 +119,13 @@ encode_mouse_button(Window *w, int button, MouseAction action, int mods) {
     if (button == GLFW_MOUSE_BUTTON_LEFT) {
         switch(action) {
             case PRESS:
-                global_state.tracked_drag_in_window = w->id; break;
+                global_state.tracked_drag_in_window = w->id;
+                global_state.tracked_drag_button = button;
+                break;
             case RELEASE:
-                global_state.tracked_drag_in_window = 0; break;
+                global_state.tracked_drag_in_window = 0;
+                global_state.tracked_drag_button = -1;
+                break;
             default:
                 break;
         }
@@ -131,12 +134,25 @@ encode_mouse_button(Window *w, int button, MouseAction action, int mods) {
 }
 
 static int
-encode_mouse_scroll(Window *w, bool upwards, int mods) {
-    return encode_mouse_event(w, upwards ? 4 : 5, PRESS, mods);
+encode_mouse_scroll(Window *w, int button, int mods) {
+    return encode_mouse_event(w, button, PRESS, mods);
 }
 
 // }}}
 
+static Window*
+window_for_id(id_type window_id) {
+    if (global_state.callback_os_window && global_state.callback_os_window->num_tabs) {
+        Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
+        for (unsigned int i = 0; i < t->num_windows; i++) {
+            Window *w = t->windows + i;
+            if (w->id == window_id) return w;
+        }
+    }
+    return window_for_window_id(window_id);
+}
+
+
 static bool
 dispatch_mouse_event(Window *w, int button, int count, int modifiers, bool grabbed) {
     bool handled = false;
@@ -220,7 +236,7 @@ cell_for_pos(Window *w, unsigned int *x, unsigned int *y, bool *in_left_half_of_
     bool in_left_half = true;
     double mouse_x = global_state.callback_os_window->mouse_x;
     double mouse_y = global_state.callback_os_window->mouse_y;
-    double left = window_left(w), top = window_top(w), right = window_right(w), bottom = window_bottom(w);
+    double left = g->left, top = g->top, right = g->right, bottom = g->bottom;
     w->mouse_pos.global_x = mouse_x - left; w->mouse_pos.global_y = mouse_y - top;
     if (clamp_to_window) {
         mouse_x = MIN(MAX(mouse_x, left), right);
@@ -271,8 +287,8 @@ do_drag_scroll(Window *w, bool upwards) {
     if (screen->linebuf == screen->main_linebuf) {
         screen_history_scroll(screen, SCROLL_LINE, upwards);
         update_drag(w);
-        if (mouse_cursor_shape != ARROW) {
-            mouse_cursor_shape = ARROW;
+        if (mouse_cursor_shape != DEFAULT_POINTER) {
+            mouse_cursor_shape = DEFAULT_POINTER;
             set_mouse_cursor(mouse_cursor_shape);
         }
         return true;
@@ -305,7 +321,16 @@ extend_selection(Window *w, bool ended, bool extend_nearest) {
 
 static void
 set_mouse_cursor_for_screen(Screen *screen) {
-    mouse_cursor_shape = screen->modes.mouse_tracking_mode == NO_TRACKING ? OPT(default_pointer_shape): OPT(pointer_shape_when_grabbed);
+    MouseShape s = screen_pointer_shape(screen);
+    if (s != INVALID_POINTER) {
+        mouse_cursor_shape = s;
+    } else {
+        if (screen->modes.mouse_tracking_mode == NO_TRACKING) {
+            mouse_cursor_shape = OPT(default_pointer_shape);
+        } else {
+            mouse_cursor_shape = OPT(pointer_shape_when_grabbed);
+        }
+    }
 }
 
 static void
@@ -323,8 +348,16 @@ handle_mouse_movement_in_kitty(Window *w, int button, bool mouse_cell_changed) {
 
 static void
 detect_url(Screen *screen, unsigned int x, unsigned int y) {
-    if (screen_detect_url(screen, x, y)) mouse_cursor_shape = HAND;
-    else set_mouse_cursor_for_screen(screen);
+    int hid = screen_detect_url(screen, x, y);
+    screen->current_hyperlink_under_mouse.id = 0;
+    if (hid != 0) {
+        mouse_cursor_shape = POINTER_POINTER;
+        if (hid > 0) {
+            screen->current_hyperlink_under_mouse.id = (hyperlink_id_type)hid;
+            screen->current_hyperlink_under_mouse.x = x;
+            screen->current_hyperlink_under_mouse.y = y;
+        }
+    } else set_mouse_cursor_for_screen(screen);
 }
 
 static bool
@@ -360,7 +393,7 @@ HANDLER(handle_move_event) {
         handle_mouse_movement_in_kitty(w, button, mouse_cell_changed | cell_half_changed);
     } else {
         if (!mouse_cell_changed && screen->modes.mouse_tracking_protocol != SGR_PIXEL_PROTOCOL) return;
-        int sz = encode_mouse_button(w, MAX(0, button), button >=0 ? DRAG : MOVE, modifiers);
+        int sz = encode_mouse_button(w, button, button >=0 ? DRAG : MOVE, modifiers);
         if (sz > 0) { mouse_event_buf[sz] = 0; write_escape_code_to_child(screen, CSI, mouse_event_buf); }
     }
 }
@@ -488,8 +521,10 @@ send_pending_click_to_window(Window *w, void *data) {
     ) {
         MousePosition current_pos = w->mouse_pos;
         w->mouse_pos = pc->mouse_pos;
+        id_type wid = w->id;
         dispatch_mouse_event(w, pc->button, pc->count, pc->modifiers, pc->grabbed);
-        w->mouse_pos = current_pos;
+        w = window_for_id(wid);
+        if (w) w->mouse_pos = current_pos;
     }
 #undef press
 }
@@ -577,16 +612,6 @@ mouse_in_region(Region *r) {
     return true;
 }
 
-static Window*
-window_for_id(id_type window_id) {
-    Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
-    for (unsigned int i = 0; i < t->num_windows; i++) {
-        Window *w = t->windows + i;
-        if (w->id == window_id) return w;
-    }
-    return NULL;
-}
-
 static Window*
 window_for_event(unsigned int *window_idx, bool *in_tab_bar) {
     Region central, tab_bar;
@@ -630,11 +655,11 @@ closest_window_for_event(unsigned int *window_idx) {
 }
 
 void
-focus_in_event() {
+focus_in_event(void) {
     // Ensure that no URL is highlighted and the mouse cursor is in default shape
     bool in_tab_bar;
     unsigned int window_idx = 0;
-    mouse_cursor_shape = BEAM;
+    mouse_cursor_shape = TEXT_POINTER;
     Window *w = window_for_event(&window_idx, &in_tab_bar);
     if (w && w->render_data.screen) {
         screen_mark_url(w->render_data.screen, 0, 0, 0, 0);
@@ -644,12 +669,26 @@ focus_in_event() {
 }
 
 void
-enter_event() {
+update_mouse_pointer_shape(void) {
+    mouse_cursor_shape = TEXT_POINTER;
+    bool in_tab_bar;
+    unsigned int window_idx = 0;
+    Window *w = window_for_event(&window_idx, &in_tab_bar);
+    if (in_tab_bar) { mouse_cursor_shape = POINTER_POINTER; }
+    else if (w && w->render_data.screen) {
+        screen_mark_url(w->render_data.screen, 0, 0, 0, 0);
+        set_mouse_cursor_for_screen(w->render_data.screen);
+    }
+    set_mouse_cursor(mouse_cursor_shape);
+}
+
+void
+enter_event(void) {
 #ifdef __APPLE__
     // On cocoa there is no way to configure the window manager to
     // focus windows on mouse enter, so we do it ourselves
     if (OPT(focus_follows_mouse) && !global_state.callback_os_window->is_focused) {
-        focus_os_window(global_state.callback_os_window, false);
+        focus_os_window(global_state.callback_os_window, false, NULL);
     }
 #endif
 }
@@ -672,6 +711,7 @@ typedef enum MouseSelectionType {
     MOUSE_SELECTION_WORD,
     MOUSE_SELECTION_LINE,
     MOUSE_SELECTION_LINE_FROM_POINT,
+    MOUSE_SELECTION_WORD_AND_LINE_FROM_POINT,
     MOUSE_SELECTION_MOVE_END,
 } MouseSelectionType;
 
@@ -703,6 +743,9 @@ mouse_selection(Window *w, int code, int button) {
         case MOUSE_SELECTION_LINE_FROM_POINT:
             if (screen_selection_range_for_line(screen, w->mouse_pos.cell_y, &start, &end) && end > w->mouse_pos.cell_x) S(EXTEND_LINE_FROM_POINT);
             break;
+        case MOUSE_SELECTION_WORD_AND_LINE_FROM_POINT:
+            if (screen_selection_range_for_line(screen, w->mouse_pos.cell_y, &start, &end) && end > w->mouse_pos.cell_x) S(EXTEND_WORD_AND_LINE_FROM_POINT);
+            break;
         case MOUSE_SELECTION_EXTEND:
             extend_selection(w, false, true);
             break;
@@ -769,7 +812,7 @@ mouse_event(const int button, int modifiers, int action) {
                         clamp_to_window = true;
                         Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
                         for (window_idx = 0; window_idx < t->num_windows && t->windows[window_idx].id != w->id; window_idx++);
-                        handle_move_event(w, button, modifiers, window_idx);
+                        handle_move_event(w, global_state.tracked_drag_button, modifiers, window_idx);
                         clamp_to_window = false;
                         debug("sent to child as drag move\n");
                         return;
@@ -792,7 +835,7 @@ mouse_event(const int button, int modifiers, int action) {
     }
     w = window_for_event(&window_idx, &in_tab_bar);
     if (in_tab_bar) {
-        mouse_cursor_shape = HAND;
+        mouse_cursor_shape = POINTER_POINTER;
         handle_tab_bar_mouse(button, modifiers, action);
         debug("handled by tab bar\n");
     } else if (w) {
@@ -811,8 +854,41 @@ mouse_event(const int button, int modifiers, int action) {
     if (mouse_cursor_shape != old_cursor) set_mouse_cursor(mouse_cursor_shape);
 }
 
+static int
+scale_scroll(MouseTrackingMode mouse_tracking_mode, double offset, bool is_high_resolution, double *pending_scroll_pixels, int cell_size) {
+// scale the scroll by the multiplier unless the mouse is grabbed. If the mouse is grabbed only change direction.
+#define SCALE_SCROLL(which) { double scale = OPT(which); if (mouse_tracking_mode) scale /= fabs(scale); offset *= scale; }
+    int s = 0;
+    if (is_high_resolution) {
+        SCALE_SCROLL(touch_scroll_multiplier);
+        double pixels = *pending_scroll_pixels + offset;
+        if (fabs(pixels) < cell_size) {
+            *pending_scroll_pixels = pixels;
+            return 0;
+        }
+        s = (int)round(pixels) / cell_size;
+        *pending_scroll_pixels = pixels - s * cell_size;
+    } else {
+        SCALE_SCROLL(wheel_scroll_multiplier);
+        s = (int) round(offset);
+        if (offset != 0) {
+            const int min_lines = mouse_tracking_mode ? 1 : OPT(wheel_scroll_min_lines);
+            if (min_lines > 0 && abs(s) < min_lines) s = offset > 0 ? min_lines : -min_lines;
+            // Always add the minimum number of lines when it is negative
+            else if (min_lines < 0) s = offset > 0 ? s - min_lines : s + min_lines;
+            // apparently on cocoa some mice generate really small yoffset values
+            // when scrolling slowly https://github.com/kovidgoyal/kitty/issues/1238
+            if (s == 0) s = offset > 0 ? 1 : -1;
+        }
+        *pending_scroll_pixels = 0;
+    }
+    return s;
+#undef SCALE_SCROLL
+}
+
 void
-scroll_event(double UNUSED xoffset, double yoffset, int flags, int modifiers) {
+scroll_event(double xoffset, double yoffset, int flags, int modifiers) {
+    debug("\x1b[36mScroll\x1b[m xoffset: %f yoffset: %f flags: %x modifiers: %s\n", xoffset, yoffset, flags, format_mods(modifiers));
     bool in_tab_bar;
     static id_type window_for_momentum_scroll = 0;
     static bool main_screen_for_momentum_scroll = false;
@@ -835,7 +911,7 @@ scroll_event(double UNUSED xoffset, double yoffset, int flags, int modifiers) {
     }
     if (!w) return;
     // Also update mouse cursor position while kitty OS window is not focused.
-    // Allows scroll events to be delivered to the child with correct pointer co-ordinates even when
+    // Allows scroll events to be delivered to the child with correct pointer coordinates even when
     // the window is not focused on macOS
     if (!osw->is_focused) {
         unsigned int x = 0, y = 0;
@@ -869,63 +945,57 @@ scroll_event(double UNUSED xoffset, double yoffset, int flags, int modifiers) {
         default:
             break;
     }
-    if (yoffset == 0.0) return;
-
     int s;
     bool is_high_resolution = flags & 1;
 
-// scale the scroll by the multiplier unless the mouse is grabbed. If the mouse is grabbed only change direction.
-#define SCALE_SCROLL(which) { double scale = OPT(which); if (screen->modes.mouse_tracking_mode) scale /= fabs(scale); yoffset *= scale; }
-    if (is_high_resolution) {
-        SCALE_SCROLL(touch_scroll_multiplier);
-        double pixels = screen->pending_scroll_pixels + yoffset;
-        if (fabs(pixels) < global_state.callback_os_window->fonts_data->cell_height) {
-            screen->pending_scroll_pixels = pixels;
-            return;
-        }
-        s = (int)round(pixels) / (int)global_state.callback_os_window->fonts_data->cell_height;
-        screen->pending_scroll_pixels = pixels - s * (int) global_state.callback_os_window->fonts_data->cell_height;
-    } else {
-        SCALE_SCROLL(wheel_scroll_multiplier);
-        s = (int) round(yoffset);
-        if (yoffset != 0) {
-            const int min_lines = screen->modes.mouse_tracking_mode ? 1 : OPT(wheel_scroll_min_lines);
-            if (min_lines > 0 && abs(s) < min_lines) s = yoffset > 0 ? min_lines : -min_lines;
-            // Always add the minimum number of lines when it is negative
-            else if (min_lines < 0) s = yoffset > 0 ? s - min_lines : s + min_lines;
-            // apparently on cocoa some mice generate really small yoffset values
-            // when scrolling slowly https://github.com/kovidgoyal/kitty/issues/1238
-            if (s == 0) s = yoffset > 0 ? 1 : -1;
-        }
-        screen->pending_scroll_pixels = 0;
-    }
-#undef SCALE_SCROLL
-    if (s == 0) return;
-    bool upwards = s > 0;
-    if (screen->modes.mouse_tracking_mode) {
-        int sz = encode_mouse_scroll(w, upwards, modifiers);
-        if (sz > 0) {
-            mouse_event_buf[sz] = 0;
-            for (s = abs(s); s > 0; s--) {
-                write_escape_code_to_child(screen, CSI, mouse_event_buf);
+    if (yoffset != 0.0) {
+        s = scale_scroll(screen->modes.mouse_tracking_mode, yoffset, is_high_resolution, &screen->pending_scroll_pixels_y, global_state.callback_os_window->fonts_data->cell_height);
+        if (s) {
+            bool upwards = s > 0;
+            if (screen->modes.mouse_tracking_mode) {
+                int sz = encode_mouse_scroll(w, upwards ? 4 : 5, modifiers);
+                if (sz > 0) {
+                    mouse_event_buf[sz] = 0;
+                    for (s = abs(s); s > 0; s--) {
+                        write_escape_code_to_child(screen, CSI, mouse_event_buf);
+                    }
+                }
+            } else {
+                if (screen->linebuf == screen->main_linebuf) screen_history_scroll(screen, abs(s), upwards);
+                else fake_scroll(w, abs(s), upwards);
             }
         }
-    } else {
-        if (screen->linebuf == screen->main_linebuf) screen_history_scroll(screen, abs(s), upwards);
-        else fake_scroll(w, abs(s), upwards);
     }
+    if (xoffset != 0.0) {
+        s = scale_scroll(screen->modes.mouse_tracking_mode, xoffset, is_high_resolution, &screen->pending_scroll_pixels_x, global_state.callback_os_window->fonts_data->cell_width);
+        if (s) {
+            if (screen->modes.mouse_tracking_mode) {
+                int sz = encode_mouse_scroll(w, s > 0 ? 6 : 7, modifiers);
+                if (sz > 0) {
+                    mouse_event_buf[sz] = 0;
+                    for (s = abs(s); s > 0; s--) {
+                        write_escape_code_to_child(screen, CSI, mouse_event_buf);
+                    }
+                }
+            }
+        }
+    }
+
 }
 
 static PyObject*
-send_mouse_event(PyObject *self UNUSED, PyObject *args) {
+send_mouse_event(PyObject *self UNUSED, PyObject *args, PyObject *kw) {
     Screen *screen;
-    unsigned int x, y;
+    int x, y, px=0, py=0, in_left_half_of_cell=0;
+
     int button, action, mods;
-    if (!PyArg_ParseTuple(args, "O!IIiii", &Screen_Type, &screen, &x, &y, &button, &action, &mods)) return NULL;
+    static const char* kwlist[] = {"screen", "cell_x", "cell_y", "button", "action", "mods", "pixel_x", "pixel_y", "in_left_half_of_cell", NULL};
+    if (!PyArg_ParseTupleAndKeywords(args, kw, "O!iiiii|iip", (char**)kwlist,
+                &Screen_Type, &screen, &x, &y, &button, &action, &mods, &px, &py, &in_left_half_of_cell)) return NULL;
 
     MouseTrackingMode mode = screen->modes.mouse_tracking_mode;
     if (mode == ANY_MODE || (mode == MOTION_MODE && action != MOVE) || (mode == BUTTON_MODE && (action == PRESS || action == RELEASE))) {
-        MousePosition mpos = {.cell_x = x, .cell_y = y};
+        MousePosition mpos = {.cell_x = x, .cell_y = y, .global_x = px, .global_y = py, .in_left_half_of_cell = in_left_half_of_cell};
         int sz = encode_mouse_event_impl(&mpos, screen->modes.mouse_tracking_protocol, button, action, mods);
         if (sz > 0) {
             mouse_event_buf[sz] = 0;
@@ -990,7 +1060,7 @@ send_mock_mouse_event_to_window(PyObject *self UNUSED, PyObject *args) {
 }
 
 static PyMethodDef module_methods[] = {
-    METHODB(send_mouse_event, METH_VARARGS),
+    {"send_mouse_event", (PyCFunction)(void (*) (void))(send_mouse_event), METH_VARARGS | METH_KEYWORDS, NULL},
     METHODB(test_encode_mouse, METH_VARARGS),
     METHODB(send_mock_mouse_event_to_window, METH_VARARGS),
     METHODB(mock_mouse_selection, METH_VARARGS),
@@ -1009,6 +1079,7 @@ init_mouse(PyObject *module) {
     PyModule_AddIntMacro(module, MOUSE_SELECTION_WORD);
     PyModule_AddIntMacro(module, MOUSE_SELECTION_LINE);
     PyModule_AddIntMacro(module, MOUSE_SELECTION_LINE_FROM_POINT);
+    PyModule_AddIntMacro(module, MOUSE_SELECTION_WORD_AND_LINE_FROM_POINT);
     PyModule_AddIntMacro(module, MOUSE_SELECTION_MOVE_END);
     if (PyModule_AddFunctions(module, module_methods) != 0) return false;
     return true;
