@@ -2,38 +2,58 @@
 # License: GPLv3 Copyright: 2021, Kovid Goyal <kovid at kovidgoyal.net>
 
 
+import enum
 import re
 import sys
+from collections import defaultdict
+from dataclasses import dataclass, fields
 from functools import lru_cache
 from typing import (
-    Any, Callable, Container, Dict, FrozenSet, Iterable, Iterator, List,
-    NamedTuple, Optional, Sequence, Tuple, Union
+    Any,
+    Callable,
+    Container,
+    Dict,
+    FrozenSet,
+    Generic,
+    Iterable,
+    Iterator,
+    List,
+    Literal,
+    NamedTuple,
+    Optional,
+    Sequence,
+    Tuple,
+    TypeVar,
+    Union,
+    get_args,
 )
 
 import kitty.fast_data_types as defines
 from kitty.conf.utils import (
-    CurrentlyParsing, KeyAction, KeyFuncWrapper, currently_parsing,
-    positive_float, positive_int, python_string, to_bool, to_cmdline, to_color,
-    uniq, unit_float
+    CurrentlyParsing,
+    KeyAction,
+    KeyFuncWrapper,
+    currently_parsing,
+    positive_float,
+    positive_int,
+    python_string,
+    to_bool,
+    to_cmdline,
+    to_color,
+    uniq,
+    unit_float,
 )
 from kitty.constants import is_macos
-from kitty.fast_data_types import (
-    CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE, Color
-)
-from kitty.fonts import FontFeature
-from kitty.key_names import (
-    character_key_name_aliases, functional_key_name_aliases,
-    get_key_name_lookup
-)
+from kitty.fast_data_types import CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE, Color, Shlex, SingleKey
+from kitty.fonts import FontFeature, FontModification, ModificationType, ModificationUnit, ModificationValue
+from kitty.key_names import character_key_name_aliases, functional_key_name_aliases, get_key_name_lookup
 from kitty.rgb import color_as_int
-from kitty.types import FloatEdges, MouseEvent, SingleKey
-from kitty.utils import expandvars, log_error, resolve_abs_or_config_path
+from kitty.types import FloatEdges, MouseEvent
+from kitty.utils import expandvars, log_error, resolve_abs_or_config_path, shlex_split
 
-KeyMap = Dict[SingleKey, str]
+KeyMap = Dict[SingleKey, List['KeyDefinition']]
 MouseMap = Dict[MouseEvent, str]
 KeySequence = Tuple[SingleKey, ...]
-SubSequenceMap = Dict[KeySequence, str]
-SequenceMap = Dict[SingleKey, SubSequenceMap]
 MINIMUM_FONT_SIZE = 4
 default_tab_separator = ' ┇'
 mod_map = {'⌃': 'CONTROL', 'CTRL': 'CONTROL', '⇧': 'SHIFT', '⌥': 'ALT', 'OPTION': 'ALT', 'OPT': 'ALT',
@@ -49,6 +69,12 @@ func_with_args = KeyFuncWrapper[FuncArgsType]()
 DELETE_ENV_VAR = '_delete_this_env_var_'
 
 
+class MapType(enum.Enum):
+    MAP = 'map'
+    MOUSE_MAP = 'mouse_map'
+    OPEN_ACTION = 'open_action'
+
+
 class InvalidMods(ValueError):
     pass
 
@@ -57,14 +83,14 @@ class InvalidMods(ValueError):
 @func_with_args(
     'pass_selection_to_program', 'new_window', 'new_tab', 'new_os_window',
     'new_window_with_cwd', 'new_tab_with_cwd', 'new_os_window_with_cwd',
-    'launch', 'mouse_handle_click'
+    'launch', 'mouse_handle_click', 'show_error',
     )
 def shlex_parse(func: str, rest: str) -> FuncArgsType:
     return func, to_cmdline(rest)
 
 
 def parse_send_text_bytes(text: str) -> bytes:
-    return python_string(text).encode('utf-8')
+    return defines.expand_ansi_c_escapes(text).encode('utf-8')
 
 
 @func_with_args('send_text')
@@ -81,6 +107,11 @@ def send_text_parse(func: str, rest: str) -> FuncArgsType:
     return func, [mode, data]
 
 
+@func_with_args('send_key')
+def send_key(func: str, rest: str) -> FuncArgsType:
+    return func, rest.split()
+
+
 @func_with_args('run_kitten', 'run_simple_kitten', 'kitten')
 def kitten_parse(func: str, rest: str) -> FuncArgsType:
     if func == 'kitten':
@@ -91,6 +122,20 @@ def kitten_parse(func: str, rest: str) -> FuncArgsType:
     return func, [args[0]] + (to_cmdline(args[1]) if len(args) > 1 else [])
 
 
+@func_with_args('open_url')
+def open_url_parse(func: str, rest: str) -> FuncArgsType:
+    from urllib.parse import urlparse
+    url = ''
+    try:
+        url = python_string(rest)
+        tokens = urlparse(url)
+        if not all((tokens.scheme, tokens.netloc,)):
+            raise ValueError('Invalid URL')
+    except Exception:
+        log_error('Ignoring invalid URL string: ' + rest)
+    return func, (url,)
+
+
 @func_with_args('goto_tab')
 def goto_tab_parse(func: str, rest: str) -> FuncArgsType:
     args = (max(0, int(rest)), )
@@ -119,7 +164,7 @@ def detach_tab_parse(func: str, rest: str) -> FuncArgsType:
     return func, (rest,)
 
 
-@func_with_args('set_background_opacity', 'goto_layout', 'toggle_layout', 'kitty_shell')
+@func_with_args('set_background_opacity', 'goto_layout', 'toggle_layout', 'kitty_shell', 'show_kitty_doc', 'set_tab_title', 'push_keyboard_mode')
 def simple_parse(func: str, rest: str) -> FuncArgsType:
     return func, [rest]
 
@@ -187,7 +232,7 @@ def paste_from_buffer(func: str, rest: str) -> FuncArgsType:
 def paste_parse(func: str, rest: str) -> FuncArgsType:
     text = ''
     try:
-        text = python_string(rest)
+        text = defines.expand_ansi_c_escapes(rest)
     except Exception:
         log_error('Ignoring invalid paste string: ' + rest)
     return func, [text]
@@ -240,8 +285,7 @@ def move_window(func: str, rest: str) -> FuncArgsType:
 
 @func_with_args('pipe')
 def pipe(func: str, rest: str) -> FuncArgsType:
-    import shlex
-    r = shlex.split(rest)
+    r = list(shlex_split(rest))
     if len(r) < 3:
         log_error('Too few arguments to pipe function')
         r = ['none', 'none', 'true']
@@ -250,8 +294,7 @@ def pipe(func: str, rest: str) -> FuncArgsType:
 
 @func_with_args('set_colors')
 def set_colors(func: str, rest: str) -> FuncArgsType:
-    import shlex
-    r = shlex.split(rest)
+    r = list(shlex_split(rest))
     if len(r) < 1:
         log_error('Too few arguments to set_colors function')
     return func, r
@@ -265,7 +308,15 @@ def remote_control(func: str, rest: str) -> FuncArgsType:
     return func, args
 
 
-@func_with_args('nth_os_window', 'nth_window', 'scroll_to_prompt', 'visual_window_select_action_trigger')
+@func_with_args('remote_control_script')
+def remote_control_script(func: str, rest: str) -> FuncArgsType:
+    func, args = shlex_parse(func, rest)
+    if len(args) < 1:
+        log_error('Too few arguments to remote_control_script function')
+    return func, args
+
+
+@func_with_args('nth_os_window', 'nth_window', 'scroll_to_prompt', 'visual_window_select_action_trigger', 'next_layout')
 def single_integer_arg(func: str, rest: str) -> FuncArgsType:
     try:
         num = int(rest)
@@ -276,6 +327,16 @@ def single_integer_arg(func: str, rest: str) -> FuncArgsType:
     return func, [num]
 
 
+@func_with_args('sleep')
+def sleep(func: str, sleep_time: str) -> FuncArgsType:
+    mult = 1
+    sleep_time = sleep_time or '1'
+    if sleep_time[-1] in 'shmd':
+        mult = {'s': 1, 'm': 60, 'h': 3600, 'd': 24 * 3600}[sleep_time[-1]]
+        sleep_time = sleep_time[:-1]
+    return func, [abs(float(sleep_time)) * mult]
+
+
 @func_with_args('disable_ligatures_in')
 def disable_ligatures_in(func: str, rest: str) -> FuncArgsType:
     parts = rest.split(maxsplit=1)
@@ -304,13 +365,13 @@ def parse_marker_spec(ftype: str, parts: Sequence[str]) -> Tuple[str, Union[str,
         if ftype.startswith('i'):
             flags |= re.IGNORECASE
         if not parts or len(parts) % 2 != 0:
-            raise ValueError('No color specified in marker: {}'.format(' '.join(parts)))
+            raise ValueError('Mark group number and text/regex are not specified in pairs: {}'.format(' '.join(parts)))
         ans = []
         for i in range(0, len(parts), 2):
             try:
                 color = max(1, min(int(parts[i]), 3))
             except Exception:
-                raise ValueError(f'color {parts[i]} in marker specification is not an integer')
+                raise ValueError(f'Mark group in marker specification is not an integer: {parts[i]}')
             sspec = parts[i + 1]
             if 'regex' not in ftype:
                 sspec = re.escape(sspec)
@@ -326,12 +387,11 @@ def parse_marker_spec(ftype: str, parts: Sequence[str]) -> Tuple[str, Union[str,
 
 @func_with_args('toggle_marker')
 def toggle_marker(func: str, rest: str) -> FuncArgsType:
-    import shlex
     parts = rest.split(maxsplit=1)
     if len(parts) != 2:
         raise ValueError(f'{rest} is not a valid marker specification')
     ftype, spec = parts
-    parts = shlex.split(spec)
+    parts = list(shlex_split(spec))
     return func, list(parse_marker_spec(ftype, parts))
 
 
@@ -363,6 +423,7 @@ def mouse_selection(func: str, rest: str) -> FuncArgsType:
             'word': defines.MOUSE_SELECTION_WORD,
             'line': defines.MOUSE_SELECTION_LINE,
             'line_from_point': defines.MOUSE_SELECTION_LINE_FROM_POINT,
+            'word_and_line_from_point': defines.MOUSE_SELECTION_WORD_AND_LINE_FROM_POINT,
         }
         setattr(mouse_selection, 'code_map', cmap)
     return func, [cmap[rest]]
@@ -370,8 +431,7 @@ def mouse_selection(func: str, rest: str) -> FuncArgsType:
 
 @func_with_args('load_config_file')
 def load_config_file(func: str, rest: str) -> FuncArgsType:
-    import shlex
-    return func, shlex.split(rest)
+    return func, list(shlex_split(rest))
 # }}}
 
 
@@ -432,26 +492,6 @@ def parse_shortcut(sc: str) -> SingleKey:
     return SingleKey(mods, is_native, key or 0)
 
 
-def adjust_line_height(x: str) -> Union[int, float]:
-    if x.endswith('%'):
-        ans = float(x[:-1].strip()) / 100.0
-        if ans < 0:
-            log_error('Percentage adjustments of cell sizes must be positive numbers')
-            return 0
-        return ans
-    return int(x)
-
-
-def adjust_baseline(x: str) -> Union[int, float]:
-    if x.endswith('%'):
-        ans = float(x[:-1].strip()) / 100.0
-        if abs(ans) > 1:
-            log_error('Percentage adjustments of the baseline cannot exceed 100%')
-            return 0
-        return ans
-    return int(x)
-
-
 def to_font_size(x: str) -> float:
     return max(MINIMUM_FONT_SIZE, float(x))
 
@@ -589,6 +629,8 @@ def optional_edge_width(x: str) -> FloatEdges:
 def hide_window_decorations(x: str) -> int:
     if x == 'titlebar-only':
         return 0b10
+    if x == 'titlebar-and-corners':
+        return 0b100
     if to_bool(x):
         return 0b01
     return 0b00
@@ -599,9 +641,16 @@ def resize_draw_strategy(x: str) -> int:
     return cmap.get(x.lower(), 0)
 
 
+def resize_debounce_time(x: str) -> Tuple[float, float]:
+    parts = x.split(maxsplit=1)
+    if len(parts) == 1:
+        return positive_float(parts[0]), 0.5
+    return positive_float(parts[0]), positive_float(parts[1])
+
+
 def visual_window_select_characters(x: str) -> str:
     import string
-    valid_characters = string.digits + string.ascii_uppercase
+    valid_characters = string.digits + string.ascii_uppercase + "-=[]\\;',./`"
     ans = x.upper()
     ans_chars = set(ans)
     if not ans_chars.issubset(set(valid_characters)):
@@ -672,16 +721,53 @@ def active_tab_title_template(x: str) -> Optional[str]:
     return None if x == 'none' else x
 
 
+class NotifyOnCmdFinish(NamedTuple):
+    when: str
+    duration: float
+    action: str
+    cmdline: Tuple[str, ...]
+
+
+def notify_on_cmd_finish(x: str) -> NotifyOnCmdFinish:
+    parts = x.split(maxsplit=3)
+    if parts[0] not in ('never', 'unfocused', 'invisible', 'always'):
+        raise ValueError(f'Unknown notify_on_cmd_finish value: {parts[0]}')
+    when = parts[0]
+    duration = 5.0
+    if len(parts) > 1:
+        duration = float(parts[1])
+    action = 'notify'
+    cmdline: Tuple[str, ...] = ()
+    if len(parts) > 2:
+        if parts[2] not in ('notify', 'bell', 'command'):
+            raise ValueError(f'Unknown notify_on_cmd_finish action: {parts[2]}')
+        action = parts[2]
+        if action == 'command':
+            if len(parts) > 3:
+                cmdline = tuple(to_cmdline(parts[3]))
+            else:
+                raise ValueError('notify_on_cmd_finish `command` action needs a command line')
+    return NotifyOnCmdFinish(when, duration, action, cmdline)
+
+
 def config_or_absolute_path(x: str, env: Optional[Dict[str, str]] = None) -> Optional[str]:
     if not x or x.lower() == 'none':
         return None
     return resolve_abs_or_config_path(x, env)
 
 
-def allow_remote_control(x: str) -> str:
-    if x != 'socket-only':
-        x = 'y' if to_bool(x) else 'n'
-    return x
+def remote_control_password(val: str, current_val: Dict[str, str]) -> Iterable[Tuple[str, Sequence[str]]]:
+    val = val.strip()
+    if val:
+        parts = to_cmdline(val, expand=False)
+        if parts[0].startswith('-'):
+            # this is done so in the future we can add --options to the cmd
+            # line of remote_control_password
+            raise ValueError('Passwords are not allowed to start with hyphens, ignoring this password')
+        if len(parts) == 1:
+            yield parts[0], tuple()
+        else:
+            yield parts[0], tuple(parts[1:])
 
 
 def clipboard_control(x: str) -> Tuple[str, ...]:
@@ -746,6 +832,10 @@ def tab_bar_margin_height(x: str) -> TabBarMarginHeight:
     return TabBarMarginHeight(next(ans), next(ans))
 
 
+def clone_source_strategies(x: str) -> FrozenSet[str]:
+    return frozenset({'venv', 'conda', 'path', 'env_var'} & set(x.lower().split(',')))
+
+
 def clear_all_mouse_actions(val: str, dict_with_parse_results: Optional[Dict[str, Any]] = None) -> bool:
     ans = to_bool(val)
     if ans and dict_with_parse_results is not None:
@@ -779,6 +869,44 @@ def font_features(val: str) -> Iterable[Tuple[str, Tuple[FontFeature, ...]]]:
         yield parts[0], tuple(features)
 
 
+def modify_font(val: str) -> Iterable[Tuple[str, FontModification]]:
+    parts = val.split()
+    pos, plen = 0, len(parts)
+    if plen < 2:
+        log_error(f"Ignoring invalid modify_font: {val}")
+        return
+    mtype: Optional[ModificationType] = getattr(ModificationType, parts[pos], None)
+    if mtype is None:
+        log_error(f"Ignoring invalid modify_font with unknown modification type: {parts[pos]}")
+        return
+    pos += 1
+    font_name = ''
+    if mtype is ModificationType.size:
+        font_name = parts[pos]
+        pos += 1
+    if plen - pos < 1:
+        log_error(f"Ignoring invalid modify_font: {val}")
+        return
+    sz = parts[pos]
+    pos += 1
+    munit = ModificationUnit.pt
+    if sz.endswith('%'):
+        munit = ModificationUnit.percent
+        sz = sz[:-1]
+    elif sz.endswith('px'):
+        munit = ModificationUnit.pixel
+        sz = sz[:-2]
+    try:
+        mvalue = float(sz)
+    except Exception:
+        log_error(f'Ignoring modify_font with invalid size: {sz}')
+        return
+    key = mtype.name
+    if font_name:
+        key += f':{font_name}'
+    yield key, FontModification(mtype, ModificationValue(mvalue, munit), font_name)
+
+
 def env(val: str, current_val: Dict[str, str]) -> Iterable[Tuple[str, str]]:
     val = val.strip()
     if val:
@@ -799,12 +927,42 @@ def store_multiple(val: str, current_val: Container[str]) -> Iterable[Tuple[str,
         yield val, val
 
 
+def menu_map(val: str, current_val: Container[str]) -> Iterable[Tuple[Tuple[str, ...], str]]:
+    parts = val.split(maxsplit=1)
+    if len(parts) != 2:
+        raise ValueError(f'Ignoring invalid menu action: {val}')
+    if parts[0] != 'global':
+        raise ValueError(f'Unknown menu type: {parts[0]}. Known types: global')
+    start = 0
+    if parts[1].startswith('"'):
+        start = 1
+        idx = parts[1].find('"', 1)
+        if idx == -1:
+            raise ValueError(f'The menu entry name in {val} must end with a double quote')
+    else:
+        idx = parts[1].find(' ')
+        if idx == -1:
+            raise ValueError(f'The menu entry {val} must have an action')
+    location = ('global',) + tuple(parts[1][start:idx].split('::'))
+    yield location, parts[1][idx+1:].lstrip()
+
+
+allowed_shell_integration_values = frozenset({'enabled', 'disabled', 'no-rc', 'no-cursor', 'no-title', 'no-prompt-mark', 'no-complete', 'no-cwd', 'no-sudo'})
+
+
 def shell_integration(x: str) -> FrozenSet[str]:
-    s = frozenset({'enabled', 'disabled', 'no-rc', 'no-cursor', 'no-title', 'no-prompt-mark', 'no-complete'})
     q = frozenset(x.lower().split())
+    if not q.issubset(allowed_shell_integration_values):
+        log_error(f'Invalid shell integration options: {q - allowed_shell_integration_values}, ignoring')
+        return q & allowed_shell_integration_values or frozenset({'invalid'})
+    return q
+
+
+def paste_actions(x: str) -> FrozenSet[str]:
+    s = frozenset({'quote-urls-at-prompt', 'confirm', 'filter', 'confirm-if-large', 'replace-dangerous-control-codes', 'replace-newline', 'no-op'})
+    q = frozenset(x.lower().split(','))
     if not q.issubset(s):
-        log_error(f'Invalid shell integration options: {q - s}, ignoring')
-        return q & s
+        raise ValueError(f'Invalid paste actions: {q - s}, ignoring')
     return q
 
 
@@ -854,7 +1012,7 @@ def narrow_symbols(val: str) -> Iterable[Tuple[Tuple[int, int], int]]:
         yield x, int(y or 1)
 
 
-def parse_key_action(action: str, action_type: str = 'map') -> KeyAction:
+def parse_key_action(action: str, action_type: MapType = MapType.MAP) -> KeyAction:
     parts = action.strip().split(maxsplit=1)
     func = parts[0]
     if len(parts) == 1:
@@ -885,7 +1043,7 @@ class AliasMap:
         self.aliases.update(aa.aliases)
 
     @lru_cache(maxsize=256)
-    def resolve_aliases(self, definition: str, map_type: str = 'map') -> Tuple[KeyAction, ...]:
+    def resolve_aliases(self, definition: str, map_type: MapType = MapType.MAP) -> Tuple[KeyAction, ...]:
         return tuple(resolve_aliases_and_parse_actions(definition, self.aliases, map_type))
 
 
@@ -901,7 +1059,7 @@ def build_action_aliases(raw: Dict[str, str], first_arg_replacement: str = '') -
 
 
 def resolve_aliases_and_parse_actions(
-    defn: str, aliases: Dict[str, List[ActionAlias]], map_type: str
+    defn: str, aliases: Dict[str, List[ActionAlias]], map_type: MapType
 ) -> Iterator[KeyAction]:
     parts = defn.split(maxsplit=1)
     if len(parts) == 1:
@@ -941,17 +1099,15 @@ def resolve_aliases_and_parse_actions(
 
 class BaseDefinition:
     no_op_actions = frozenset(('noop', 'no-op', 'no_op'))
-    map_type: str = 'map'
+    map_type: MapType = MapType.MAP
     definition_location: CurrentlyParsing
 
     def __init__(self, definition: str = '') -> None:
+        if definition in BaseDefinition.no_op_actions:
+            definition = ''
         self.definition = definition
         self.definition_location = currently_parsing.__copy__()
 
-    @property
-    def is_no_op(self) -> bool:
-        return self.definition in self.no_op_actions
-
     def pretty_repr(self, *fields: str) -> str:
         kwds = []
         defaults = self.__class__()
@@ -964,8 +1120,12 @@ class BaseDefinition:
         return f'{self.__class__.__name__}({", ".join(kwds)})'
 
 
+def resolve_key_mods(kitty_mod: int, mods: int) -> int:
+    return SingleKey(mods=mods).resolve_kitty_mod(kitty_mod).mods
+
+
 class MouseMapping(BaseDefinition):
-    map_type: str = 'mouse_map'
+    map_type: MapType = MapType.MOUSE_MAP
 
     def __init__(
         self, button: int = 0, mods: int = 0, repeat_count: int = 1, grabbed: bool = False,
@@ -982,7 +1142,7 @@ class MouseMapping(BaseDefinition):
 
     def resolve_and_copy(self, kitty_mod: int) -> 'MouseMapping':
         ans = MouseMapping(
-            self.button, defines.resolve_key_mods(kitty_mod, self.mods), self.repeat_count, self.grabbed,
+            self.button, resolve_key_mods(kitty_mod, self.mods), self.repeat_count, self.grabbed,
             self.definition)
         ans.definition_location = self.definition_location
         return ans
@@ -992,39 +1152,139 @@ class MouseMapping(BaseDefinition):
         return MouseEvent(self.button, self.mods, self.repeat_count, self.grabbed)
 
 
+T = TypeVar('T')
+
+
+class LiteralField(Generic[T]):
+    def __init__(self, vals: Tuple[T, ...]):
+        self._vals = vals
+
+    def __set_name__(self, owner: object, name: str) -> None:
+        self._name = "_" + name
+
+    def __get__(self, obj: object, type: Optional[type] = None) -> T:
+        if obj is None:
+            return self._vals[0]
+        return getattr(obj, self._name, self._vals[0])
+
+    def __set__(self, obj: object, value: str) -> None:
+        if value not in self._vals:
+            raise KeyError(f'Invalid value for {self._name[1:]}: {value!r}')
+        object.__setattr__(obj, self._name, value)
+
+
+OnUnknown = Literal['beep', 'end', 'ignore', 'passthrough']
+OnAction = Literal['keep', 'end']
+
+
+@dataclass(init=False, frozen=True)
+class KeyMapOptions:
+    when_focus_on: str = ''
+    new_mode: str = ''
+    mode: str = ''
+    on_unknown: LiteralField[OnUnknown] = LiteralField[OnUnknown](get_args(OnUnknown))
+    on_action: LiteralField[OnAction] = LiteralField[OnAction](get_args(OnAction))
+
+
+default_key_map_options = KeyMapOptions()
+allowed_key_map_options = frozenset(f.name for f in fields(KeyMapOptions))
+
+
 class KeyDefinition(BaseDefinition):
 
     def __init__(
         self, is_sequence: bool = False, trigger: SingleKey = SingleKey(),
-            rest: Tuple[SingleKey, ...] = (), definition: str = ''
+        rest: Tuple[SingleKey, ...] = (), definition: str = '',
+        options: KeyMapOptions = default_key_map_options
     ):
         super().__init__(definition)
         self.is_sequence = is_sequence
         self.trigger = trigger
         self.rest = rest
+        self.options = options
+
+    @property
+    def is_suitable_for_global_shortcut(self) -> bool:
+        return not self.options.when_focus_on and not self.options.mode and not self.options.new_mode and not self.is_sequence
 
     def __repr__(self) -> str:
-        return self.pretty_repr('is_sequence', 'trigger', 'rest')
+        return self.pretty_repr('is_sequence', 'trigger', 'rest', 'options')
+
+    def human_repr(self) -> str:
+        ans = self.definition or 'no-op'
+        if self.options.when_focus_on:
+            ans = f'[--when-focus-on={self.options.when_focus_on}]{ans}'
+        return ans
+
+    def shift_sequence_and_copy(self) -> 'KeyDefinition':
+        return KeyDefinition(self.is_sequence, self.trigger, self.rest[1:], self.definition, self.options)
 
     def resolve_and_copy(self, kitty_mod: int) -> 'KeyDefinition':
         def r(k: SingleKey) -> SingleKey:
-            mods = defines.resolve_key_mods(kitty_mod, k.mods)
-            return k._replace(mods=mods)
+            return k.resolve_kitty_mod(kitty_mod)
         ans = KeyDefinition(
             self.is_sequence, r(self.trigger), tuple(map(r, self.rest)),
-            self.definition
+            self.definition, self.options
         )
         ans.definition_location = self.definition_location
         return ans
 
 
+class KeyboardMode:
+
+    on_unknown: OnUnknown = get_args(OnUnknown)[0]
+    on_action : OnAction = get_args(OnAction)[0]
+    is_sequence: bool = False
+
+    def __init__(self, name: str = '') -> None:
+        self.name = name
+        self.keymap: KeyMap = defaultdict(list)
+
+
+KeyboardModeMap = Dict[str, KeyboardMode]
+
+
+def parse_options_for_map(val: str) -> Tuple[KeyMapOptions, str]:
+    expecting_arg = ''
+    ans = KeyMapOptions()
+    s = Shlex(val)
+    while (tok := s.next_word())[0] > -1:
+        x = tok[1]
+        if expecting_arg:
+            object.__setattr__(ans, expecting_arg, x)
+            expecting_arg = ''
+        elif x.startswith('--'):
+            expecting_arg = x[2:]
+            k, sep, v = expecting_arg.partition('=')
+            k = k.replace('-', '_')
+            expecting_arg = k
+            if expecting_arg not in allowed_key_map_options:
+                raise KeyError(f'The map option {x} is unknown. Allowed options: {", ".join(allowed_key_map_options)}')
+            if sep == '=':
+                object.__setattr__(ans, k, v)
+                expecting_arg = ''
+        else:
+            return ans, val[tok[0]:]
+    return ans, ''
+
+
 def parse_map(val: str) -> Iterable[KeyDefinition]:
     parts = val.split(maxsplit=1)
-    if len(parts) != 2:
-        return
-    sc, action = parts
+    options = default_key_map_options
+    if len(parts) == 2:
+        sc, action = parts
+        if sc.startswith('--'):
+            options, leftover = parse_options_for_map(val)
+            parts = leftover.split(maxsplit=1)
+            if len(parts) == 1:
+                sc, action = parts[0], ''
+            else:
+                sc = parts[0]
+                action = ' '.join(parts[1:])
+    else:
+        sc, action = val, ''
     sc, action = sc.strip().strip(sequence_sep), action.strip()
-    if not sc or not action:
+    if not sc:
         return
     is_sequence = sequence_sep in sc
     if is_sequence:
@@ -1055,18 +1315,22 @@ def parse_map(val: str) -> Iterable[KeyDefinition]:
             return
     if is_sequence:
         if trigger is not None:
-            yield KeyDefinition(True, trigger, rest, definition=action)
+            yield KeyDefinition(True, trigger, rest, definition=action, options=options)
     else:
         assert key is not None
-        yield KeyDefinition(False, SingleKey(mods, is_native, key), definition=action)
+        yield KeyDefinition(False, SingleKey(mods, is_native, key), definition=action, options=options)
 
 
 def parse_mouse_map(val: str) -> Iterable[MouseMapping]:
     parts = val.split(maxsplit=3)
-    if len(parts) != 4:
+    if len(parts) == 4:
+        xbutton, event, modes, action = parts
+    elif len(parts) > 2:
+        xbutton, event, modes = parts
+        action = ''
+    else:
         log_error(f'Ignoring invalid mouse action: {val}')
         return
-    xbutton, event, modes, action = parts
     kparts = xbutton.split('+')
     if len(kparts) > 1:
         mparts, obutton = kparts[:-1], kparts[-1].lower()
@@ -1135,3 +1399,16 @@ def deprecated_send_text(key: str, val: str, ans: Dict[str, Any]) -> None:
     key_str = f'{sc} send_text {mode} {text}'
     for k in parse_map(key_str):
         ans['map'].append(k)
+
+
+def deprecated_adjust_line_height(key: str, x: str, opts_dict: Dict[str, Any]) -> None:
+    fm = {'adjust_line_height': 'cell_height', 'adjust_baseline': 'baseline', 'adjust_column_width': 'cell_width'}[key]
+    mtype = getattr(ModificationType, fm)
+    if x.endswith('%'):
+        ans = float(x[:-1].strip())
+        if ans < 0:
+            log_error(f'Percentage adjustments of {key} must be positive numbers')
+            return
+        opts_dict['modify_font'][fm] = FontModification(mtype, ModificationValue(ans, ModificationUnit.percent))
+    else:
+        opts_dict['modify_font'][fm] = FontModification(mtype, ModificationValue(int(x), ModificationUnit.pixel))
